package ru.dbotthepony.mc.otm.container

import com.google.common.collect.ImmutableSet
import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import net.minecraft.world.item.ItemStack

data class ItemFilterSet(val filter: ImmutableSet<ItemFilter>, val isWhitelist: Boolean = false) : ItemFilter {
	constructor(list: Collection<ItemFilter>, isWhitelist: Boolean = false) : this(ImmutableSet.copyOf(list), isWhitelist)

	val size: Int
		get() = filter.size

	override val allowAll: Boolean by lazy(LazyThreadSafetyMode.PUBLICATION) {
		filter.isEmpty() && !isWhitelist || filter.isNotEmpty() && isWhitelist && filter.any { it.allowAll }
	}

	override val denyAll: Boolean by lazy(LazyThreadSafetyMode.PUBLICATION) {
		filter.isEmpty() && isWhitelist || filter.isNotEmpty() && !isWhitelist && filter.any { it.denyAll }
	}

	fun replace(index: Int, value: ItemFilter): ItemFilterSet {
		if (index !in filter.indices || value in filter)
			return this

		val values = ObjectArrayList(filter)
		values[index] = value
		return copy(filter = ImmutableSet.copyOf(values))
	}

	fun addOrReplace(index: Int, value: ItemFilter): ItemFilterSet {
		if (index !in filter.indices)
			return add(value)
		else
			return replace(index, value)
	}

	fun add(value: ItemFilter): ItemFilterSet {
		if (value in filter)
			return this

		val values = ObjectArrayList(filter)
		values.add(value)
		return copy(filter = ImmutableSet.copyOf(values))
	}

	fun removeAt(index: Int): ItemFilterSet {
		if (index !in filter.indices)
			return this

		if (filter.size == 1)
			return copy(filter = ImmutableSet.of())

		val values = ObjectArrayList(filter)
		values.removeAt(index)
		return copy(filter = ImmutableSet.copyOf(values))
	}

	fun indexOf(value: ItemFilter): Int {
		return filter.asList().indexOf(value)
	}

	operator fun get(index: Int): ItemFilter {
		return filter.asList().getOrElse(index) { ItemFilter.EMPTY }
	}

	fun clear(): ItemFilterSet {
		if (filter.isEmpty())
			return this

		return copy(filter = ImmutableSet.of())
	}

	fun isWhitelist(flag: Boolean): ItemFilterSet {
		if (flag == isWhitelist)
			return this
		else
			return copy(isWhitelist = flag)
	}

	override fun test(value: ItemStack): Boolean {
		return if (denyAll || value.isEmpty)
			false
		else if (allowAll)
			true
		else if (filter.any { it.test(value) })
			isWhitelist
		else
			!isWhitelist
	}

	override val type: ItemFilter.Type<*>
		get() = Companion

	override val depth: Int by lazy {
		if (filter.isNotEmpty())
			return@lazy 1 + filter.maxOf { it.depth }

		return@lazy 1
	}

	// TODO: can not be "lazy" cached because that will break with /reload command
	override val displayItems: ImmutableSet<ItemStackKey> get() {
		val sub = filter.map { it.displayItems }
		val results = ArrayList<ItemStackKey>(sub.sumOf { it.size })
		sub.forEach { results.addAll(it) }
		return ImmutableSet.copyOf(results)
	}

	companion object : ItemFilter.Type<ItemFilterSet> {
		override val codec: MapCodec<ItemFilterSet> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					Codec.list(ItemFilter.CODEC, 0, 40).fieldOf("filter").forGetter { ObjectArrayList(it.filter) },
					Codec.BOOL.optionalFieldOf("isWhitelist", false).forGetter { it.isWhitelist },
				).apply(it, ::ItemFilterSet)
			}
		}

		val EMPTY = ItemFilterSet(ImmutableSet.of())

		val CODEC: Codec<ItemFilterSet> by lazy {
			codec.codec()
		}
	}
}
