package ru.dbotthepony.mc.otm.data.world

import com.mojang.datafixers.util.Either
import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.util.RandomSource
import net.neoforged.bus.api.IEventBus
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.util.math.Decimal
import ru.dbotthepony.mc.otm.util.nextDecimal
import ru.dbotthepony.mc.otm.data.codec.DecimalCodec
import ru.dbotthepony.mc.otm.registry.MBuiltInRegistries
import ru.dbotthepony.mc.otm.registry.MDeferredRegister
import ru.dbotthepony.mc.otm.registry.MRegistries

fun interface SampledDecimal {
	fun sample(source: RandomSource): Decimal
}

abstract class DecimalProvider : SampledDecimal {
	interface Type<T : DecimalProvider> {
		val codec: MapCodec<T>
	}

	abstract val minValue: Decimal
	abstract val maxValue: Decimal
	abstract val type: Type<*>

	companion object {
		val CODEC: Codec<DecimalProvider> by lazy {
			Codec
				.either(DecimalCodec, MBuiltInRegistries.DECIMAL_PROVIDER_TYPE.byNameCodec().dispatch({ it.type }, { it.codec }))
				.xmap(
					{ c -> c.map(::ConstantDecimal, { it }) },
					{ if (it.type === ConstantDecimal) Either.left(it.minValue) else Either.right(it) }
				)
		}

		private val registrar = MDeferredRegister(MRegistries.DECIMAL_PROVIDER_TYPE, OverdriveThatMatters.MOD_ID)

		init {
			registrar.register("zero") { ConstantDecimal.Zero }
			registrar.register("constant") { ConstantDecimal }
			registrar.register("uniform") { UniformDecimal }
		}

		internal fun register(bus: IEventBus) {
			registrar.register(bus)
		}
	}
}

class ConstantDecimal(val value: Decimal) : DecimalProvider() {
	object Zero : DecimalProvider(), Type<Zero> {
		override fun sample(source: RandomSource): Decimal {
			return Decimal.ZERO
		}

		override val codec: MapCodec<Zero> = MapCodec.unit(this)

		override val minValue: Decimal
			get() = Decimal.ZERO
		override val maxValue: Decimal
			get() = Decimal.ZERO
		override val type: Type<*>
			get() = this
	}

	override fun sample(source: RandomSource): Decimal {
		return value
	}

	override val minValue: Decimal
		get() = value
	override val maxValue: Decimal
		get() = value
	override val type: Type<*>
		get() = Companion

	companion object : Type<ConstantDecimal> {
		override val codec: MapCodec<ConstantDecimal> = RecordCodecBuilder.mapCodec {
			it.group(DecimalCodec.fieldOf("value").forGetter(ConstantDecimal::value)).apply(it, ::ConstantDecimal)
		}
	}
}

class UniformDecimal(override val minValue: Decimal, override val maxValue: Decimal, val round: Boolean = true) : DecimalProvider() {
	constructor(range: ClosedRange<Decimal>, round: Boolean = true) : this(range.start, range.endInclusive, round)

	init {
		require(minValue <= maxValue) { "Invalid range provided: $minValue .. $maxValue" }
	}

	override fun sample(source: RandomSource): Decimal {
		return source.nextDecimal(minValue, maxValue, round = round)
	}

	override val type: Type<*>
		get() = Companion

	companion object : Type<UniformDecimal> {
		override val codec: MapCodec<UniformDecimal> = RecordCodecBuilder.mapCodec {
			it.group(
				DecimalCodec.fieldOf("minValue").forGetter(UniformDecimal::minValue),
				DecimalCodec.fieldOf("maxValue").forGetter(UniformDecimal::maxValue),
				Codec.BOOL.optionalFieldOf("round", true).forGetter(UniformDecimal::round)
			).apply(it, ::UniformDecimal)
		}
	}
}
